#include <cstdio>
#include <cstring>
#include <iconv.h>
#include <unistd.h>
#include <pqxx/pqxx>
#include <ft2build.h>
#include FT_FREETYPE_H
#include <tinyptc.h>
#include "flagtrigger.h"
#include "widestring.h"
#include "fetch_current_tournament.h"

int my_draw_text(const widestring &str, unsigned char *buf, int xpos, int ypos, bool real_render, int r, int g, int b, std::vector<FT_Face> &fontlist);

Tournament active_tournament;
std::vector<FT_Face> fonts;

void init(pqxx::connection &conn)
{
	conn.perform(FetchCurrentTournament(&active_tournament));

	if (active_tournament.id == -1) {
		std::fprintf(stderr, "No active tournament\n");
	} else {
		std::fprintf(stderr, "Current tournament is %d (name: '%s')\n",
			active_tournament.id, active_tournament.name.c_str());
	}
}

unsigned char framebuf[800 * 600 * 4];

void main_loop(pqxx::connection &conn)
{
	if (active_tournament.id == -1) {
		// No active tournament, sleep a second or so and exit
		sleep(1);
		return;
	}

	memset(framebuf, 0, 800*600*4);
	
	pqxx::work t(conn, "trx");

	// fetch all songs
	pqxx::result res( t.exec("SELECT * FROM songs WHERE title LIKE 'M%'") );
	unsigned y = 0;
	for (pqxx::result::const_iterator i = res.begin(); i != res.end(); ++i) {
		my_draw_text(i["title"].as(widestring()), framebuf, 0, y, 1, 255, 255, 255, fonts);
		y += 20;
//		std::fprintf(stderr, "%s\n", i["title"].c_str());
	}
	t.commit();

	ptc_update(framebuf);
	sleep(1);
}

void init_freetype()
{
	FT_Library library;
	FT_Face face;
	if (FT_Init_FreeType(&library))
		throw std::runtime_error("FreeType init failed.");

	// Georgia
	if (FT_New_Face(library, "/usr/share/fonts/truetype/msttcorefonts/Georgia.ttf", 0, &face))
		throw std::runtime_error("Face opening failed.");
	if (FT_Set_Char_Size(face, 0, 12 * 64, 96, 96))
		throw std::runtime_error("Size set failed.");
	fonts.push_back(face);

	// FreeSerif
	if (FT_New_Face(library, "/usr/share/fonts/truetype/freefont/FreeSerif.ttf", 0, &face)) {
		std::fprintf(stderr, "Warning: Couldn't open FreeSerif, some glyphs might not be available\n");
	} else {
		if (FT_Set_Char_Size(face, 0, 12 * 64, 96, 96))
			throw std::runtime_error("Size set failed.");
		fonts.push_back(face);
	}
	
	// Arial Unicode MS
	if (FT_New_Face(library, "arialuni.ttf", 0, &face)) {
		std::fprintf(stderr, "Warning: Couldn't open Arial Unicode MS, some glyphs might not be available\n");
	} else {
		if (FT_Set_Char_Size(face, 0, 12 * 64, 96, 96))
			throw std::runtime_error("Size set failed.");
		fonts.push_back(face);
	}
}

int my_draw_text(const widestring &str, unsigned char *buf, int xpos, int ypos, bool real_render, int r, int g, int b, std::vector<FT_Face> &fontlist)
{
	FT_GlyphSlot slot;
	int x = 0;

	for (widestring::const_iterator i = str.begin(); i != str.end(); ++i) {
		int glyph_index;
		for (std::vector<FT_Face>::const_iterator j = fontlist.begin(); j != fontlist.end(); ++j) {
			glyph_index = FT_Get_Char_Index(*j, *i);
			if (glyph_index == 0)
				continue;

			if (FT_Load_Glyph(*j, glyph_index, FT_LOAD_RENDER))
				throw std::runtime_error("Couldn't load glyph");
			slot = (*j)->glyph;
			break;
		}
		if (glyph_index == 0) {
			std::fprintf(stderr, "Warning: Could not find a glyph in any font for U+%x, ignoring\n", *i);
			continue;
		}

		if (real_render) {
			int y;
			FT_Bitmap *bm = &(slot->bitmap);
			for (y = 0; y < bm->rows; y++) {
				int xx;
				int dsty = ypos - slot->bitmap_top + y;
				if (dsty < 0 || dsty > 599) continue;

				unsigned char *dst = buf + dsty * 800*4 + (x + xpos + slot->bitmap_left)*4;
				unsigned char *src = bm->buffer + y * bm->width;
				for (xx = 0; xx < bm->width; xx++) {
					*dst = (*dst * (256-*src) + r * *src) >> 8;
					*dst++;
					*dst = (*dst * (256-*src) + g * *src) >> 8;
					*dst++;
					*dst = (*dst * (256-*src) + b * *src) >> 8;
					*dst++;
					*dst++ = 0;
					src++;
				}
			}
		}

		x += slot->advance.x >> 6;
	}

	return x;
}


int main(int argc, char **argv)
{
	ptc_open("CCBS bigscreen", 800, 600);
	
	try {
		init_freetype();
		pqxx::connection conn("dbname=ccbs host=altersex.samfundet.no user=ccbs password=GeT|>>B_");
		FlagTrigger tournament_changed(conn, "active_tournament");
		FlagTrigger rounds_changed(conn, "active_groups");
		
		// when active_tournament or active_rounds is changed, we destroy everything and start from scratch
		// (at least currently)
		for ( ;; ) {
			tournament_changed.reset_flag();
			rounds_changed.reset_flag();
			init(conn);
			do {
				main_loop(conn);
				conn.get_notifs();
			} while (!tournament_changed.get_flag() && !rounds_changed.get_flag());
			std::fprintf(stderr, "active_tournament or active_groups changed, resetting...\n");
		}
	} catch (const std::exception &e) {
		std::fprintf(stderr, "Exception: %s\n", e.what());
		exit(1);
	}
	
	return 0;
}
